#include <FL/Fl.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_File_Browser.H>
#include <FL/Fl_File_Icon.H>
#include <FL/Fl_Input.H>
#include <FL/Fl_PNG_Image.H>
#include <FL/Fl_Toggle_Button.H>
#include <FL/Fl_Double_Window.H>
#include <FL/filename.H>

#include <iostream>
#include <string>
#include <stdlib.h>
#include <unistd.h>

/* Tango icons
 * for f in *.png; do xxd -i $f >> icons.h; done
 */
#include "icons.h"

#define DOUBLE_CLICK_TIMEOUT 1.0

static Fl_Double_Window *win;
static Fl_File_Browser *br;
static Fl_Box *box;
static Fl_Button *bt_up;
static Fl_Input *input;
static std::string current_dir = "/", selection = "", selection_bn = "";
static int selection_n = 0;
static bool show_hidden = false;

static Fl_PNG_Image go_up(NULL, go_up_png, go_up_png_len);
static Fl_PNG_Image go_up_gray(NULL, go_up_gray_png, go_up_gray_png_len);

static void up_callback(Fl_Widget *);
static void top_callback(Fl_Widget *);
static void home_callback(Fl_Widget *);
static void hidden_callback(Fl_Widget *);
static void selection_timeout(void);
static void br_callback(Fl_Widget *);
static void br_change_dir(void);
static void close_cb(Fl_Widget *);


static void up_callback(Fl_Widget *)
{
  if (current_dir != "/") {
    current_dir.erase(current_dir.length() - 1);
    current_dir.erase(current_dir.find_last_of("/") + 1);
    br_change_dir();
  }
}

static void top_callback(Fl_Widget *) {
  current_dir = "/";
  br_change_dir();
}

static void home_callback(Fl_Widget *)
{
  char *home = fl_getenv("HOME");

  if (home) {
    current_dir = std::string(home);
    if (current_dir[current_dir.length()] != '/') {
      current_dir += "/";
    }
    br_change_dir();
  }
}

static void hidden_callback(Fl_Widget *) {
  show_hidden = show_hidden ? false : true;
  br_change_dir();
}

static void selection_timeout(void) {
  selection_n = 0;
}

static void br_callback(Fl_Widget *)
{
  /* some workaround to change directories on double-click */
  if (selection_n == 0) {
    selection_n = br->value();
    Fl::add_timeout(DOUBLE_CLICK_TIMEOUT, (Fl_Timeout_Handler)selection_timeout);
  } else {
    Fl::remove_timeout((Fl_Timeout_Handler)selection_timeout);
    if (br->value() == selection_n) {
      selection_n = 0;
      std::string fname = current_dir + br->text(br->value());
      if (fname[fname.length()-1] == '/' && access(fname.c_str(), R_OK) == 0) {
        /* double-clicked on directory */
        current_dir = fname;
        br_change_dir();
      } else {
        /* double-clicked on file */
        close_cb(NULL);
      }
    } else {
      selection_n = br->value();
      Fl::add_timeout(DOUBLE_CLICK_TIMEOUT, (Fl_Timeout_Handler)selection_timeout);
    }
  }

  selection = current_dir;

  if (br->value() != 0) {
    selection += br->text(br->value());
  }

  selection_bn = selection;

  /* just like basename() */
  if (selection_bn != "/") {
    if (selection_bn[selection_bn.length() - 1] == '/') {
      selection_bn.erase(selection_bn.length() - 1);
    }
    selection_bn.erase(0, selection_bn.find_last_of("/") + 1);
  }

  input->value(selection_bn.c_str());
}

static void br_change_dir(void)
{
  selection = selection_bn = "";
  selection_n = 0;

  br->load(current_dir.c_str());
  br->remove(1); /* remove ".." entry */
  box->label(current_dir.c_str());

  if (!show_hidden) {
    /* filter() doesn't affect hidden directories */
    for (int i = br->size(); i > 0; i--) {
      if (br->text(i)[0] == '.') {
        br->remove(i);
      }
    }
  }

  if (current_dir == "/" && bt_up->active()) {
    bt_up->deactivate();
    bt_up->image(go_up_gray);
  } else if (current_dir != "/" && !bt_up->active()) {
    bt_up->activate();
    bt_up->image(go_up);
  }
}

static void close_cb(Fl_Widget *) {
  if (selection != "") {
    std::cout << selection << std::endl;
  }
  win->hide();
}

int file_browser(void)
{
  Fl_File_Icon *icon_any = new Fl_File_Icon("*", Fl_File_Icon::ANY);
  Fl_File_Icon *icon_dir = new Fl_File_Icon("*", Fl_File_Icon::DIRECTORY);
  icon_any->load_png(text_x_generic_png, text_x_generic_png_len);
  icon_dir->load_png(folder_png, folder_png_len);

  int w = 800;
  int h = 600;

  Fl_Group *g, *g_top;

  win = new Fl_Double_Window(w, h, "Open File");
  {
    g = new Fl_Group(0, 0, w, h);
    {
      g_top = new Fl_Group(0, 0, w, 40);
      {
        int bt_w = 36;

        box = new Fl_Box(10, 10, w - bt_w*4 - 24, 20, "/");
        box->align(FL_ALIGN_INSIDE|FL_ALIGN_TOP_LEFT);
        //box->box(FL_FLAT_BOX);
        //box->color(FL_WHITE);

        bt_up = new Fl_Button(w - bt_w*4 - 10, 5, bt_w, 30);
        bt_up->tooltip("Parent Directory");
        bt_up->image(go_up_gray);
        bt_up->deactivate();
        bt_up->callback(up_callback);
        bt_up->clear_visible_focus();

       { Fl_Button *o = new Fl_Button(w - bt_w*3 - 10, 5, bt_w, 30);
        o->tooltip("Top Directory");
        o->image(new Fl_PNG_Image(NULL, go_top_png, go_top_png_len));
        o->callback(top_callback);
        o->clear_visible_focus(); }

       { Fl_Button *o = new Fl_Button(w - bt_w*2 - 10, 5, bt_w, 30);
        o->tooltip("Home Directory");
        o->image(new Fl_PNG_Image(NULL, go_home_png, go_home_png_len));
        o->callback(home_callback);
        o->clear_visible_focus(); }

       { Fl_Toggle_Button *o = new Fl_Toggle_Button(w - bt_w - 10, 5, bt_w, 30, "show\nhidden");
        o->labelsize(8);
        o->tooltip("Show Hidden Elements");
        o->callback(hidden_callback);
        o->clear_visible_focus(); }
      }
      g_top->resizable(box);
      g_top->end();

      br = new Fl_File_Browser(10, 40, w - 20, h - g_top->h() - 46);
      br->type(FL_HOLD_BROWSER);
      br->callback(br_callback);
      br_change_dir();

      input = new Fl_Input(10, br->y() + br->h() + 10, w - 20, 26);
    }
    g->resizable(br);
    g->end();
  }
  win->end();
  win->resizable(g);
  /* center window */
  win->position((Fl::w() - win->w()) / 2, (Fl::h() - win->h()) / 2);
  win->callback(close_cb);
  win->show();

  return Fl::run();
}

int main(void) {
  return file_browser();
}

